"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getComposerFromJSONSchema = getComposerFromJSONSchema;
const graphql_1 = require("graphql");
const graphql_compose_1 = require("graphql-compose");
const graphql_scalars_1 = require("graphql-scalars");
const json_machete_1 = require("json-machete");
const utils_1 = require("@graphql-mesh/utils");
const directives_js_1 = require("./directives.js");
const getJSONSchemaStringFormatScalarMap_js_1 = require("./getJSONSchemaStringFormatScalarMap.js");
const getUnionTypeComposers_js_1 = require("./getUnionTypeComposers.js");
const getValidTypeName_js_1 = require("./getValidTypeName.js");
const scalars_js_1 = require("./scalars.js");
const formatScalarMapWithoutAjv = {
    byte: graphql_scalars_1.GraphQLByte,
    binary: scalars_js_1.GraphQLFile,
    'date-time': graphql_scalars_1.GraphQLDateTime,
    time: graphql_scalars_1.GraphQLTime,
    email: graphql_scalars_1.GraphQLEmailAddress,
    ipv4: graphql_scalars_1.GraphQLIPv4,
    ipv6: graphql_scalars_1.GraphQLIPv6,
    uri: graphql_scalars_1.GraphQLURL,
    uuid: graphql_scalars_1.GraphQLUUID,
    'unix-time': graphql_scalars_1.GraphQLTimestamp,
    int64: graphql_scalars_1.GraphQLBigInt,
    int32: graphql_1.GraphQLInt,
    decimal: graphql_1.GraphQLFloat,
    float: graphql_1.GraphQLFloat,
};
const deepMergedInputTypeComposerFields = new WeakMap();
function deepMergeInputTypeComposerFields(existingInputTypeComposer, newInputTypeComposer) {
    let mergedInputTypes = deepMergedInputTypeComposerFields.get(existingInputTypeComposer);
    if (!mergedInputTypes) {
        mergedInputTypes = new WeakSet();
        deepMergedInputTypeComposerFields.set(existingInputTypeComposer, mergedInputTypes);
    }
    if (mergedInputTypes.has(newInputTypeComposer)) {
        return;
    }
    mergedInputTypes.add(newInputTypeComposer);
    const existingInputTypeComposerFields = existingInputTypeComposer.getFields();
    const newInputTypeComposerFields = newInputTypeComposer.getFields();
    for (const [newFieldKey, newFieldValue] of Object.entries(newInputTypeComposerFields)) {
        const existingFieldValue = existingInputTypeComposerFields[newFieldKey];
        if (!existingFieldValue) {
            existingInputTypeComposerFields[newFieldKey] = newFieldValue;
        }
        else {
            const existingFieldUnwrappedTC = typeof existingFieldValue.type?.getUnwrappedTC === 'function'
                ? existingFieldValue.type?.getUnwrappedTC()
                : undefined;
            const newFieldUnwrappedTC = typeof newFieldValue.type.getUnwrappedTC === 'function'
                ? newFieldValue.type.getUnwrappedTC()
                : undefined;
            if (existingFieldUnwrappedTC instanceof graphql_compose_1.InputTypeComposer &&
                newFieldUnwrappedTC instanceof graphql_compose_1.InputTypeComposer) {
                deepMergeInputTypeComposerFields(existingFieldUnwrappedTC, newFieldUnwrappedTC);
            }
            else {
                existingInputTypeComposerFields[newFieldKey] = newFieldValue;
            }
        }
    }
}
function deepMergeObjectTypeComposerFields(existingObjectTypeComposer, newObjectTypeComposer) {
    const existingObjectTypeComposerFields = existingObjectTypeComposer.getFields();
    const newObjectTypeComposerFields = newObjectTypeComposer.getFields();
    for (const [newFieldKey, newFieldValue] of Object.entries(newObjectTypeComposerFields)) {
        const existingFieldValue = existingObjectTypeComposerFields[newFieldKey];
        if (!existingFieldValue) {
            existingObjectTypeComposerFields[newFieldKey] = newFieldValue;
        }
        else {
            const existingFieldUnwrappedTC = typeof existingFieldValue.type?.getUnwrappedTC === 'function'
                ? existingFieldValue.type?.getUnwrappedTC()
                : undefined;
            const newFieldUnwrappedTC = typeof newFieldValue.type.getUnwrappedTC === 'function'
                ? newFieldValue.type.getUnwrappedTC()
                : undefined;
            if (existingFieldUnwrappedTC instanceof graphql_compose_1.ObjectTypeComposer &&
                newFieldUnwrappedTC instanceof graphql_compose_1.ObjectTypeComposer) {
                deepMergeObjectTypeComposerFields(existingFieldUnwrappedTC, newFieldUnwrappedTC);
            }
            else {
                if (newFieldUnwrappedTC &&
                    existingFieldUnwrappedTC &&
                    !isUnspecificType(newFieldUnwrappedTC) &&
                    isUnspecificType(existingFieldUnwrappedTC)) {
                    continue;
                }
                existingObjectTypeComposerFields[newFieldKey] = newFieldValue;
            }
        }
    }
}
function getComposerFromJSONSchema({ subgraphName, schema, logger, getScalarForFormat, }) {
    const schemaComposer = new graphql_compose_1.SchemaComposer();
    const formatScalarMap = (0, getJSONSchemaStringFormatScalarMap_js_1.getJSONSchemaStringFormatScalarMap)();
    const getDefaultScalarForFormat = (format) => formatScalarMapWithoutAjv[format] || formatScalarMap.get(format);
    const rootInputTypeNameComposerMap = {
        QueryInput: () => schemaComposer.Query,
        MutationInput: () => schemaComposer.Mutation,
        SubscriptionInput: () => schemaComposer.Subscription,
    };
    return (0, json_machete_1.visitJSONSchema)(schema, {
        enter(subSchema, { path, visitedSubschemaResultMap }) {
            if (typeof subSchema === 'boolean' || subSchema.title === 'Any') {
                const typeComposer = schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON);
                return subSchema
                    ? {
                        input: typeComposer,
                        output: typeComposer,
                    }
                    : undefined;
            }
            if (!subSchema) {
                throw new Error(`Something is wrong with ${path}`);
            }
            if (subSchema.type === 'array') {
                if (subSchema.items != null &&
                    typeof subSchema.items === 'object' &&
                    Object.keys(subSchema.items).length > 0) {
                    return {
                        // These are filled after enter
                        get input() {
                            const typeComposers = visitedSubschemaResultMap.get(subSchema.items);
                            return typeComposers.input.getTypePlural();
                        },
                        get output() {
                            const typeComposers = visitedSubschemaResultMap.get(subSchema.items);
                            return typeComposers.output.getTypePlural();
                        },
                        ...subSchema,
                    };
                }
                if (subSchema.contains) {
                    // Scalars cannot be in union type
                    const typeComposer = schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON).getTypePlural();
                    return {
                        input: typeComposer,
                        output: typeComposer,
                        nullable: subSchema.nullable,
                        readOnly: subSchema.readOnly,
                        writeOnly: subSchema.writeOnly,
                        default: subSchema.default,
                        deprecated: subSchema.deprecated,
                    };
                }
                // If it doesn't have any clue
                {
                    // const typeComposer = getGenericJSONScalar({
                    //   schemaComposer,
                    //   isInput: false,
                    //   subSchema,
                    //   validateWithJSONSchema,
                    // }).getTypePlural();
                    const typeComposer = schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON).getTypePlural();
                    return {
                        input: typeComposer,
                        output: typeComposer,
                        description: subSchema.description,
                        nullable: subSchema.nullable,
                        readOnly: subSchema.readOnly,
                        writeOnly: subSchema.writeOnly,
                        default: subSchema.default,
                        deprecated: subSchema.deprecated,
                    };
                }
            }
            if (subSchema.pattern) {
                let typeScriptType;
                switch (subSchema.type) {
                    case 'number':
                        typeScriptType = 'number';
                        break;
                    case 'integer':
                        if (subSchema.format === 'int64') {
                            typeScriptType = 'bigint';
                        }
                        else {
                            typeScriptType = 'number';
                        }
                        break;
                    default:
                        typeScriptType = 'string';
                        break;
                }
                schemaComposer.addDirective(directives_js_1.RegExpDirective);
                schemaComposer.addDirective(directives_js_1.TypeScriptDirective);
                const typeComposer = schemaComposer.createScalarTC({
                    name: (0, getValidTypeName_js_1.getValidTypeName)({
                        schemaComposer,
                        isInput: false,
                        subSchema,
                    }),
                    directives: [
                        {
                            name: 'regexp',
                            args: {
                                subgraph: subgraphName,
                                pattern: subSchema.pattern,
                            },
                        },
                        {
                            name: 'typescript',
                            args: {
                                subgraph: subgraphName,
                                type: typeScriptType,
                            },
                        },
                    ],
                });
                return {
                    input: typeComposer,
                    output: typeComposer,
                    nullable: subSchema.nullable,
                    readOnly: subSchema.readOnly,
                    writeOnly: subSchema.writeOnly,
                    deprecated: subSchema.deprecated,
                };
            }
            if (subSchema.const) {
                const scalarTypeName = (0, getValidTypeName_js_1.getValidTypeName)({
                    schemaComposer,
                    isInput: false,
                    subSchema,
                });
                schemaComposer.addDirective(directives_js_1.EnumDirective);
                schemaComposer.addDirective(directives_js_1.TypeScriptDirective);
                schemaComposer.addDirective(directives_js_1.ExampleDirective);
                let enumValueName = (0, utils_1.sanitizeNameForGraphQL)(subSchema.const.toString());
                if (enumValueName === 'true' || enumValueName === 'false') {
                    enumValueName = enumValueName.toUpperCase();
                }
                const typeComposer = schemaComposer.createEnumTC({
                    name: scalarTypeName,
                    values: {
                        [enumValueName]: {
                            directives: [
                                {
                                    name: 'enum',
                                    args: {
                                        subgraph: subgraphName,
                                        value: JSON.stringify(subSchema.const),
                                    },
                                },
                            ],
                        },
                    },
                    directives: [
                        {
                            name: 'typescript',
                            args: {
                                subgraph: subgraphName,
                                type: JSON.stringify(subSchema.const),
                            },
                        },
                        {
                            name: 'example',
                            args: {
                                subgraph: subgraphName,
                                value: subSchema.const,
                            },
                        },
                    ],
                    extensions: {
                        default: subSchema.const,
                    },
                });
                return {
                    input: typeComposer,
                    output: typeComposer,
                    nullable: subSchema.nullable,
                    readOnly: subSchema.readOnly,
                    writeOnly: subSchema.writeOnly,
                    deprecated: subSchema.deprecated,
                };
            }
            if (subSchema.enum && subSchema.type !== 'boolean') {
                const values = {};
                for (const value of subSchema.enum) {
                    let enumKey = (0, utils_1.sanitizeNameForGraphQL)(value.toString());
                    if (enumKey === 'false' || enumKey === 'true' || enumKey === 'null') {
                        enumKey = enumKey.toUpperCase();
                    }
                    if (typeof enumKey === 'string' && enumKey.length === 0) {
                        enumKey = '_';
                    }
                    schemaComposer.addDirective(directives_js_1.EnumDirective);
                    // Falsy values are ignored by GraphQL
                    // eslint-disable-next-line no-unneeded-ternary
                    const enumValue = value || value === 0 ? value : value?.toString();
                    const directives = [];
                    if (enumValue !== enumKey) {
                        directives.push({
                            name: 'enum',
                            args: {
                                subgraph: subgraphName,
                                value: JSON.stringify(enumValue),
                            },
                        });
                    }
                    values[enumKey] = {
                        directives,
                        value: enumValue,
                    };
                }
                const directives = [];
                if (subSchema.examples?.length) {
                    schemaComposer.addDirective(directives_js_1.ExampleDirective);
                    for (const example of subSchema.examples) {
                        directives.push({
                            name: 'example',
                            args: {
                                subgraph: subgraphName,
                                value: example,
                            },
                        });
                    }
                }
                const typeComposer = schemaComposer.createEnumTC({
                    name: (0, getValidTypeName_js_1.getValidTypeName)({
                        schemaComposer,
                        isInput: false,
                        subSchema,
                    }),
                    values,
                    description: subSchema.description,
                    directives,
                    extensions: {
                        default: subSchema.default,
                    },
                });
                return {
                    input: typeComposer,
                    output: typeComposer,
                    nullable: subSchema.nullable,
                    readOnly: subSchema.readOnly,
                    writeOnly: subSchema.writeOnly,
                    default: subSchema.default,
                    deprecated: subSchema.deprecated,
                };
            }
            if (Array.isArray(subSchema.type)) {
                const validTypes = subSchema.type.filter((typeName) => typeName !== 'null');
                if (validTypes.length === 1) {
                    subSchema.type = validTypes[0];
                    // continue with the single type
                }
                else {
                    const typeComposer = schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON);
                    return {
                        input: typeComposer,
                        output: typeComposer,
                        nullable: subSchema.nullable,
                        readOnly: subSchema.readOnly,
                        writeOnly: subSchema.writeOnly,
                        default: subSchema.default,
                        deprecated: subSchema.deprecated,
                    };
                }
            }
            if (subSchema.format) {
                const formatScalar = getScalarForFormat?.(subSchema.format) || getDefaultScalarForFormat(subSchema.format);
                if (formatScalar) {
                    const typeComposer = schemaComposer.getAnyTC(formatScalar);
                    return {
                        input: typeComposer,
                        output: typeComposer,
                        description: subSchema.description,
                        nullable: subSchema.nullable,
                        readOnly: subSchema.readOnly,
                        writeOnly: subSchema.writeOnly,
                        default: subSchema.default,
                        deprecated: subSchema.deprecated,
                    };
                }
            }
            if (subSchema.minimum === 0) {
                const typeComposer = schemaComposer.getAnyTC(subSchema.type === 'integer' ? graphql_scalars_1.GraphQLNonNegativeInt : graphql_scalars_1.GraphQLNonNegativeFloat);
                return {
                    input: typeComposer,
                    output: typeComposer,
                    description: subSchema.description,
                    nullable: subSchema.nullable,
                    readOnly: subSchema.readOnly,
                    writeOnly: subSchema.writeOnly,
                    default: subSchema.default,
                    deprecated: subSchema.deprecated,
                };
            }
            else if (subSchema.minimum > 0) {
                const typeComposer = schemaComposer.getAnyTC(subSchema.type === 'integer' ? graphql_scalars_1.GraphQLPositiveInt : graphql_scalars_1.GraphQLPositiveFloat);
                return {
                    input: typeComposer,
                    output: typeComposer,
                    description: subSchema.description,
                    nullable: subSchema.nullable,
                    readOnly: subSchema.readOnly,
                    writeOnly: subSchema.writeOnly,
                    default: subSchema.default,
                    deprecated: subSchema.deprecated,
                };
            }
            if (subSchema.maximum === 0) {
                const typeComposer = schemaComposer.getAnyTC(subSchema.type === 'integer' ? graphql_scalars_1.GraphQLNonPositiveInt : graphql_scalars_1.GraphQLNonPositiveFloat);
                return {
                    input: typeComposer,
                    output: typeComposer,
                    description: subSchema.description,
                    nullable: subSchema.nullable,
                    readOnly: subSchema.readOnly,
                    writeOnly: subSchema.writeOnly,
                    default: subSchema.default,
                    deprecated: subSchema.deprecated,
                };
            }
            else if (subSchema.maximum < 0) {
                const typeComposer = schemaComposer.getAnyTC(subSchema.type === 'integer' ? graphql_scalars_1.GraphQLNegativeInt : graphql_scalars_1.GraphQLNegativeFloat);
                return {
                    input: typeComposer,
                    output: typeComposer,
                    description: subSchema.description,
                    nullable: subSchema.nullable,
                    readOnly: subSchema.readOnly,
                    writeOnly: subSchema.writeOnly,
                    default: subSchema.default,
                    deprecated: subSchema.deprecated,
                };
            }
            if (subSchema.maximum > Number.MAX_SAFE_INTEGER ||
                subSchema.minimum < Number.MIN_SAFE_INTEGER) {
                const typeComposer = schemaComposer.getAnyTC(graphql_scalars_1.GraphQLBigInt);
                return {
                    input: typeComposer,
                    output: typeComposer,
                    description: subSchema.description,
                    nullable: subSchema.nullable,
                    readOnly: subSchema.readOnly,
                    writeOnly: subSchema.writeOnly,
                    default: subSchema.default,
                    deprecated: subSchema.deprecated,
                };
            }
            switch (subSchema.type) {
                case 'boolean': {
                    const typeComposer = schemaComposer.getAnyTC(graphql_1.GraphQLBoolean);
                    return {
                        input: typeComposer,
                        output: typeComposer,
                        description: subSchema.description,
                        nullable: subSchema.nullable,
                        readOnly: subSchema.readOnly,
                        writeOnly: subSchema.writeOnly,
                        default: subSchema.default,
                        deprecated: subSchema.deprecated,
                    };
                }
                case 'null': {
                    const typeComposer = schemaComposer.getAnyTC(scalars_js_1.GraphQLVoid);
                    return {
                        input: typeComposer,
                        output: typeComposer,
                        description: subSchema.description,
                        nullable: subSchema.nullable,
                        readOnly: subSchema.readOnly,
                        writeOnly: subSchema.writeOnly,
                        default: subSchema.default,
                        deprecated: subSchema.deprecated,
                    };
                }
                case 'integer': {
                    const typeComposer = schemaComposer.getAnyTC(graphql_1.GraphQLInt);
                    return {
                        input: typeComposer,
                        output: typeComposer,
                        description: subSchema.description,
                        nullable: subSchema.nullable,
                        readOnly: subSchema.readOnly,
                        writeOnly: subSchema.writeOnly,
                        default: subSchema.default,
                        deprecated: subSchema.deprecated,
                    };
                }
                case 'number': {
                    const typeComposer = schemaComposer.getAnyTC(graphql_1.GraphQLFloat);
                    return {
                        input: typeComposer,
                        output: typeComposer,
                        description: subSchema.description,
                        nullable: subSchema.nullable,
                        readOnly: subSchema.readOnly,
                        writeOnly: subSchema.writeOnly,
                        default: subSchema.default,
                        deprecated: subSchema.deprecated,
                    };
                }
                case 'string': {
                    if (subSchema.minLength === 1 && subSchema.maxLength == null) {
                        const tc = schemaComposer.getAnyTC(graphql_scalars_1.GraphQLNonEmptyString);
                        return {
                            input: tc,
                            output: tc,
                            description: subSchema.description,
                            nullable: subSchema.nullable,
                            readOnly: subSchema.readOnly,
                            writeOnly: subSchema.writeOnly,
                            default: subSchema.default,
                            deprecated: subSchema.deprecated,
                        };
                    }
                    if (subSchema.minLength != null || subSchema.maxLength != null) {
                        schemaComposer.addDirective(directives_js_1.LengthDirective);
                        const lengthArgs = {
                            subgraph: subgraphName,
                        };
                        if (subSchema.minLength != null) {
                            lengthArgs.min = subSchema.minLength;
                        }
                        if (subSchema.maxLength != null) {
                            lengthArgs.max = subSchema.maxLength;
                        }
                        const typeComposer = schemaComposer.createScalarTC({
                            name: (0, getValidTypeName_js_1.getValidTypeName)({
                                schemaComposer,
                                isInput: false,
                                subSchema,
                            }),
                            description: subSchema.description,
                            directives: [
                                {
                                    name: 'length',
                                    args: lengthArgs,
                                },
                            ],
                        });
                        return {
                            input: typeComposer,
                            output: typeComposer,
                            description: subSchema.description,
                            nullable: subSchema.nullable,
                            readOnly: subSchema.readOnly,
                            writeOnly: subSchema.writeOnly,
                            default: subSchema.default,
                            deprecated: subSchema.deprecated,
                        };
                    }
                    const typeComposer = schemaComposer.getAnyTC(graphql_1.GraphQLString);
                    return {
                        input: typeComposer,
                        output: typeComposer,
                        description: subSchema.description,
                        nullable: subSchema.nullable,
                        readOnly: subSchema.readOnly,
                        writeOnly: subSchema.writeOnly,
                        default: subSchema.default,
                        deprecated: subSchema.deprecated,
                    };
                }
                case 'object': {
                    switch (subSchema.title) {
                        case '_schema':
                            return {
                                output: schemaComposer,
                                ...subSchema,
                            };
                        case 'Query':
                            return {
                                output: schemaComposer.Query,
                                ...subSchema,
                            };
                        case 'Mutation':
                            return {
                                output: schemaComposer.Mutation,
                                ...subSchema,
                            };
                        case 'Subscription': {
                            if (path === '/properties/subscription') {
                                return {
                                    output: schemaComposer.Subscription,
                                    ...subSchema,
                                };
                            }
                            subSchema.title = 'Subscription_';
                            break;
                        }
                    }
                }
            }
            if (subSchema.oneOf && !subSchema.properties) {
                schemaComposer.addDirective(directives_js_1.OneOfDirective);
                const input = schemaComposer.createInputTC({
                    name: (0, getValidTypeName_js_1.getValidTypeName)({
                        schemaComposer,
                        isInput: true,
                        subSchema,
                    }),
                    fields: {},
                    directives: [
                        {
                            name: 'oneOf',
                            args: {
                                subgraph: subgraphName,
                            },
                        },
                    ],
                });
                const extensions = {};
                if (subSchema.$comment?.startsWith('statusCodeOneOfIndexMap:')) {
                    const statusCodeOneOfIndexMapStr = subSchema.$comment.replace('statusCodeOneOfIndexMap:', '');
                    const statusCodeOneOfIndexMap = JSON.parse(statusCodeOneOfIndexMapStr);
                    if (statusCodeOneOfIndexMap) {
                        extensions.statusCodeOneOfIndexMap = statusCodeOneOfIndexMap;
                    }
                }
                const output = schemaComposer.createUnionTC({
                    name: (0, getValidTypeName_js_1.getValidTypeName)({
                        schemaComposer,
                        isInput: false,
                        subSchema,
                    }),
                    description: subSchema.description,
                    types: [],
                    extensions,
                });
                return {
                    input,
                    output,
                    ...subSchema,
                };
            }
            if (subSchema.properties ||
                subSchema.allOf ||
                subSchema.anyOf ||
                subSchema.additionalProperties) {
                if (subSchema.title === 'Any') {
                    const typeComposer = schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON);
                    return {
                        input: typeComposer,
                        output: typeComposer,
                        description: subSchema.description,
                        nullable: subSchema.nullable,
                        readOnly: subSchema.readOnly,
                        writeOnly: subSchema.writeOnly,
                        default: subSchema.default,
                        deprecated: subSchema.deprecated,
                    };
                }
                const config = {
                    name: (0, getValidTypeName_js_1.getValidTypeName)({
                        schemaComposer,
                        isInput: false,
                        subSchema,
                    }),
                    description: subSchema.description,
                    fields: {},
                    directives: [],
                    extensions: {
                        default: subSchema.default,
                    },
                };
                if (subSchema.examples?.length) {
                    schemaComposer.addDirective(directives_js_1.ExampleDirective);
                    for (const example of subSchema.examples) {
                        config.directives.push({
                            name: 'example',
                            args: {
                                subgraph: subgraphName,
                                value: example,
                            },
                        });
                    }
                }
                if (subSchema.discriminator?.propertyName) {
                    schemaComposer.addDirective(directives_js_1.DiscriminatorDirective);
                }
                const directives = [];
                if (subSchema.examples?.length) {
                    schemaComposer.addDirective(directives_js_1.ExampleDirective);
                    for (const example of subSchema.examples) {
                        directives.push({
                            name: 'example',
                            args: {
                                subgraph: subgraphName,
                                value: example,
                            },
                        });
                    }
                }
                return {
                    input: schemaComposer.createInputTC({
                        name: (0, getValidTypeName_js_1.getValidTypeName)({
                            schemaComposer,
                            isInput: true,
                            subSchema,
                        }),
                        description: subSchema.description,
                        fields: {},
                        directives,
                        extensions: {
                            default: subSchema.default,
                        },
                    }),
                    output: subSchema.discriminator
                        ? schemaComposer.createInterfaceTC(config)
                        : schemaComposer.createObjectTC(config),
                    ...subSchema,
                    ...(subSchema.properties ? { properties: { ...subSchema.properties } } : {}),
                    ...(subSchema.allOf ? { allOf: [...subSchema.allOf] } : {}),
                    ...(subSchema.additionalProperties
                        ? {
                            additionalProperties: subSchema.additionalProperties === true
                                ? true
                                : { ...subSchema.additionalProperties },
                        }
                        : {}),
                    ...(subSchema.discriminatorMapping
                        ? { discriminatorMapping: { ...subSchema.discriminatorMapping } }
                        : {}),
                };
            }
            return subSchema;
        },
        leave(subSchemaAndTypeComposers, { path }) {
            // const validateWithJSONSchema = getValidateFnForSchemaPath(ajv, path, schema);
            const subSchemaOnly = {
                ...subSchemaAndTypeComposers,
                input: undefined,
                output: undefined,
            };
            if (subSchemaOnly.discriminator?.propertyName) {
                schemaComposer.addDirective(directives_js_1.DiscriminatorDirective);
                const discriminatorArgs = {
                    subgraph: subgraphName,
                    field: subSchemaOnly.discriminator.propertyName,
                };
                if (subSchemaOnly.discriminator.mapping) {
                    const mappingByName = {};
                    for (const discriminatorValue in subSchemaOnly.discriminatorMapping) {
                        const discType = subSchemaOnly.discriminatorMapping[discriminatorValue];
                        mappingByName[discriminatorValue] = discType.output.getTypeName();
                    }
                    discriminatorArgs.mapping = Object.entries(mappingByName);
                }
                subSchemaAndTypeComposers.output.setDirectiveByName('discriminator', discriminatorArgs);
            }
            if (subSchemaAndTypeComposers.oneOf && !subSchemaAndTypeComposers.properties) {
                const isPlural = subSchemaAndTypeComposers.oneOf.some(({ output }) => output instanceof graphql_compose_1.ListComposer);
                if (isPlural) {
                    const { input, output, flatten } = (0, getUnionTypeComposers_js_1.getUnionTypeComposers)({
                        subgraphName,
                        schemaComposer,
                        typeComposersList: subSchemaAndTypeComposers.oneOf.map(({ input, output }) => ({
                            input: input.ofType || input,
                            output: output.ofType || output,
                        })),
                        subSchemaAndTypeComposers,
                        logger,
                    });
                    return {
                        input: input.getTypePlural(),
                        output: output instanceof graphql_compose_1.ListComposer
                            ? output
                            : output.getTypePlural(),
                        nullable: subSchemaAndTypeComposers.nullable,
                        default: subSchemaAndTypeComposers.default,
                        readOnly: subSchemaAndTypeComposers.readOnly,
                        writeOnly: subSchemaAndTypeComposers.writeOnly,
                        flatten,
                        deprecated: subSchemaAndTypeComposers.deprecated,
                    };
                }
                return (0, getUnionTypeComposers_js_1.getUnionTypeComposers)({
                    subgraphName,
                    schemaComposer,
                    typeComposersList: subSchemaAndTypeComposers.oneOf,
                    subSchemaAndTypeComposers,
                    logger,
                });
            }
            const fieldMap = {};
            const inputFieldMap = {};
            let isList = false;
            if (subSchemaAndTypeComposers.allOf) {
                let ableToUseGraphQLInputObjectType = true;
                for (const maybeTypeComposers of subSchemaAndTypeComposers.allOf) {
                    let { input: inputTypeComposer, output: outputTypeComposer } = maybeTypeComposers;
                    if (inputTypeComposer instanceof graphql_compose_1.ListComposer) {
                        isList = true;
                        inputTypeComposer = inputTypeComposer.ofType;
                    }
                    if (outputTypeComposer instanceof graphql_compose_1.ListComposer) {
                        isList = true;
                        outputTypeComposer = outputTypeComposer.ofType;
                    }
                    if (inputTypeComposer instanceof graphql_compose_1.ScalarTypeComposer ||
                        inputTypeComposer instanceof graphql_compose_1.EnumTypeComposer) {
                        ableToUseGraphQLInputObjectType = false;
                    }
                    else {
                        const inputTypeElemFieldMap = inputTypeComposer.getFields();
                        for (const fieldName in inputTypeElemFieldMap) {
                            const newInputField = inputTypeElemFieldMap[fieldName];
                            const existingInputField = inputFieldMap[fieldName];
                            if (!existingInputField) {
                                inputFieldMap[fieldName] = newInputField;
                            }
                            else {
                                /*
                                  If the new field collides with an existing field:
                
                                    - If both the existing and the new field have an input type composer, combine their subfields
                                    - Otherwise, replace the existing field with the new one
                                */
                                const existingInputFieldUnwrappedTC = typeof existingInputField.type?.getUnwrappedTC === 'function'
                                    ? existingInputField.type.getUnwrappedTC()
                                    : undefined;
                                const newInputFieldUnwrappedTC = typeof newInputField.type?.getUnwrappedTC === 'function'
                                    ? newInputField.type.getUnwrappedTC()
                                    : undefined;
                                if (existingInputFieldUnwrappedTC instanceof graphql_compose_1.InputTypeComposer &&
                                    newInputFieldUnwrappedTC instanceof graphql_compose_1.InputTypeComposer) {
                                    deepMergeInputTypeComposerFields(existingInputFieldUnwrappedTC, newInputFieldUnwrappedTC);
                                }
                                else {
                                    inputFieldMap[fieldName] = newInputField;
                                }
                            }
                        }
                    }
                    if ((0, graphql_compose_1.isSomeInputTypeComposer)(outputTypeComposer)) {
                        schemaComposer.addDirective(directives_js_1.ResolveRootDirective);
                        fieldMap[outputTypeComposer.getTypeName()] = {
                            type: outputTypeComposer,
                            directives: [
                                {
                                    name: 'resolveRoot',
                                    args: {
                                        subgraph: subgraphName,
                                    },
                                },
                            ],
                        };
                    }
                    else if (outputTypeComposer instanceof graphql_compose_1.UnionTypeComposer) {
                        const outputTCElems = outputTypeComposer.getTypes();
                        for (const outputTCElem of outputTCElems) {
                            const outputTypeElemFieldMap = outputTCElem.getFields();
                            for (const fieldName in outputTypeElemFieldMap) {
                                const field = outputTypeElemFieldMap[fieldName];
                                fieldMap[fieldName] = field;
                            }
                        }
                    }
                    else {
                        if (outputTypeComposer instanceof graphql_compose_1.InterfaceTypeComposer) {
                            subSchemaAndTypeComposers.output.addInterface(outputTypeComposer);
                            schemaComposer.addSchemaMustHaveType(subSchemaAndTypeComposers.output);
                        }
                        const typeElemFieldMap = outputTypeComposer.getFields();
                        for (const fieldName in typeElemFieldMap) {
                            const newField = typeElemFieldMap[fieldName];
                            const existingField = fieldMap[fieldName];
                            if (!existingField) {
                                fieldMap[fieldName] = newField;
                            }
                            else {
                                /*
                                  If the new field collides with an existing field:
                
                                    - If both the existing and the new field have an object type composer, combine their subfields
                                    - Otherwise, replace the existing field with the new one
                                */
                                const existingFieldUnwrappedTC = typeof existingField.type?.getUnwrappedTC === 'function'
                                    ? existingField.type.getUnwrappedTC()
                                    : undefined;
                                const newFieldUnwrappedTC = typeof newField.type?.getUnwrappedTC === 'function'
                                    ? newField.type.getUnwrappedTC()
                                    : undefined;
                                if (existingFieldUnwrappedTC instanceof graphql_compose_1.ObjectTypeComposer &&
                                    newFieldUnwrappedTC instanceof graphql_compose_1.ObjectTypeComposer) {
                                    deepMergeObjectTypeComposerFields(existingFieldUnwrappedTC, newFieldUnwrappedTC);
                                }
                                else {
                                    if (newFieldUnwrappedTC &&
                                        existingFieldUnwrappedTC &&
                                        !isUnspecificType(newFieldUnwrappedTC) &&
                                        isUnspecificType(existingFieldUnwrappedTC)) {
                                        continue;
                                    }
                                    fieldMap[fieldName] = newField;
                                }
                            }
                        }
                    }
                }
                if (subSchemaAndTypeComposers.examples?.length) {
                    schemaComposer.addDirective(directives_js_1.ExampleDirective);
                    const directives = subSchemaAndTypeComposers.output.getDirectives() || [];
                    for (const example of subSchemaAndTypeComposers.examples) {
                        directives.push({
                            name: 'example',
                            args: {
                                subgraph: subgraphName,
                                value: example,
                            },
                        });
                    }
                    subSchemaAndTypeComposers.output.setDirectives(directives);
                }
                subSchemaAndTypeComposers.output.addFields(fieldMap);
                subSchemaAndTypeComposers.output.setExtensions({
                    // validateWithJSONSchema,
                    default: subSchemaAndTypeComposers.default,
                });
                if (ableToUseGraphQLInputObjectType) {
                    subSchemaAndTypeComposers.input.addFields(inputFieldMap);
                    if (subSchemaAndTypeComposers.examples?.length) {
                        schemaComposer.addDirective(directives_js_1.ExampleDirective);
                        const directives = subSchemaAndTypeComposers.input.getDirectives() || [];
                        for (const example of subSchemaAndTypeComposers.examples) {
                            directives.push({
                                name: 'example',
                                args: {
                                    subgraph: subgraphName,
                                    value: example,
                                },
                            });
                        }
                        subSchemaAndTypeComposers.input.setDirectives(directives);
                    }
                    subSchemaAndTypeComposers.input.setExtensions({
                        default: subSchemaAndTypeComposers.default,
                    });
                }
                else {
                    subSchemaAndTypeComposers.input = schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON);
                }
            }
            if (subSchemaAndTypeComposers.anyOf) {
                // It should not have `required` because it is `anyOf` not `allOf`
                let ableToUseGraphQLInputObjectType = true;
                for (const typeComposers of subSchemaAndTypeComposers.anyOf) {
                    let { input: inputTypeComposer, output: outputTypeComposer } = typeComposers;
                    if (inputTypeComposer instanceof graphql_compose_1.ListComposer ||
                        outputTypeComposer instanceof graphql_compose_1.ListComposer) {
                        isList = true;
                        inputTypeComposer = inputTypeComposer.ofType;
                        outputTypeComposer = outputTypeComposer.ofType;
                    }
                    if (inputTypeComposer instanceof graphql_compose_1.ScalarTypeComposer ||
                        inputTypeComposer instanceof graphql_compose_1.EnumTypeComposer) {
                        ableToUseGraphQLInputObjectType = false;
                    }
                    else {
                        const inputTypeElemFieldMap = inputTypeComposer.getFields();
                        for (const fieldName in inputTypeElemFieldMap) {
                            // In case of conflict set it to JSON
                            // TODO: But instead we can convert that field into a oneOf of all possible types
                            if (inputFieldMap[fieldName]) {
                                let existingType = inputFieldMap[fieldName].type;
                                if (typeof existingType === 'function') {
                                    existingType = existingType();
                                }
                                let newType = inputTypeElemFieldMap[fieldName].type;
                                if (typeof newType === 'function') {
                                    newType = newType();
                                }
                                const newTypeName = newType.getTypeName().replace('!', '');
                                const existingTypeName = existingType.getTypeName().replace('!', '');
                                if (existingTypeName !== newTypeName) {
                                    if (newTypeName !== 'JSON') {
                                        inputFieldMap[fieldName] = {
                                            type: schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON),
                                        };
                                    }
                                    if (existingTypeName === 'JSON') {
                                        const field = inputTypeElemFieldMap[fieldName];
                                        inputFieldMap[fieldName] = (0, graphql_1.isNonNullType)(field.type.getType())
                                            ? {
                                                ...field,
                                                type: () => field.type.ofType,
                                            }
                                            : field;
                                    }
                                }
                            }
                            else {
                                const field = inputTypeElemFieldMap[fieldName];
                                inputFieldMap[fieldName] = (0, graphql_1.isNonNullType)(field.type.getType())
                                    ? {
                                        ...field,
                                        type: () => field.type.ofType,
                                    }
                                    : field;
                            }
                        }
                    }
                    if (outputTypeComposer instanceof graphql_compose_1.ScalarTypeComposer) {
                        const typeName = outputTypeComposer.getTypeName();
                        // In case of conflict set it to JSON
                        // TODO: But instead we can convert that field into a union of all possible types
                        if (fieldMap[typeName]) {
                            const existingTypeName = fieldMap[typeName]?.type?.getTypeName();
                            if (existingTypeName === 'JSON') {
                                schemaComposer.addDirective(directives_js_1.ResolveRootDirective);
                                fieldMap[typeName] = {
                                    type: outputTypeComposer,
                                    directives: [
                                        {
                                            name: 'resolveRoot',
                                            args: {
                                                subgraph: subgraphName,
                                            },
                                        },
                                    ],
                                };
                            }
                            if (typeName !== 'JSON' && existingTypeName !== typeName) {
                                fieldMap[typeName] = {
                                    type: schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON),
                                };
                            }
                        }
                        else {
                            schemaComposer.addDirective(directives_js_1.ResolveRootDirective);
                            fieldMap[typeName] = {
                                type: outputTypeComposer,
                                directives: [
                                    {
                                        name: 'resolveRoot',
                                        args: {
                                            subgraph: subgraphName,
                                        },
                                    },
                                ],
                            };
                        }
                    }
                    else if (outputTypeComposer instanceof graphql_compose_1.ObjectTypeComposer ||
                        outputTypeComposer instanceof graphql_compose_1.InterfaceTypeComposer) {
                        const typeElemFieldMap = outputTypeComposer.getFields();
                        for (const fieldName in typeElemFieldMap) {
                            // In case of conflict set it to JSON
                            // TODO: But instead we can convert that field into a union of all possible types
                            const field = typeElemFieldMap[fieldName];
                            const existingField = fieldMap[fieldName];
                            fieldMap[fieldName] = {
                                ...field,
                                type: () => {
                                    if (!field.type) {
                                    }
                                    const fieldType = field.type.getType();
                                    const namedType = (0, graphql_1.getNamedType)(fieldType);
                                    if (existingField) {
                                        const existingFieldType = existingField.type();
                                        const existingNamedType = (0, graphql_1.getNamedType)(existingFieldType);
                                        const existingTypeName = existingNamedType.name;
                                        const newTypeName = namedType.name;
                                        if (existingTypeName !== 'JSON' && existingNamedType.name !== namedType.name) {
                                            return schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON);
                                        }
                                        if (newTypeName === 'JSON') {
                                            return existingFieldType;
                                        }
                                    }
                                    return field.type.getType();
                                },
                            };
                        }
                    }
                }
                let outputTypeComposer = subSchemaAndTypeComposers.output;
                if ('ofType' in outputTypeComposer) {
                    outputTypeComposer = outputTypeComposer.ofType;
                }
                outputTypeComposer.addFields(fieldMap);
                if (subSchemaAndTypeComposers.examples?.length) {
                    schemaComposer.addDirective(directives_js_1.ExampleDirective);
                    const directives = outputTypeComposer.getDirectives() || [];
                    for (const example of subSchemaAndTypeComposers.examples) {
                        directives.push({
                            name: 'example',
                            args: {
                                subgraph: subgraphName,
                                value: example,
                            },
                        });
                    }
                    outputTypeComposer.setDirectives(directives);
                }
                outputTypeComposer.setExtensions({
                    // validateWithJSONSchema,
                    default: subSchemaAndTypeComposers.default,
                });
                let inputTypeComposer = subSchemaAndTypeComposers.input;
                if ('ofType' in inputTypeComposer) {
                    inputTypeComposer = inputTypeComposer.ofType;
                }
                if (ableToUseGraphQLInputObjectType) {
                    inputTypeComposer.addFields(inputFieldMap);
                    if (subSchemaAndTypeComposers.examples?.length) {
                        schemaComposer.addDirective(directives_js_1.ExampleDirective);
                        const directives = inputTypeComposer.getDirectives() || [];
                        for (const example of subSchemaAndTypeComposers.examples) {
                            directives.push({
                                name: 'example',
                                args: {
                                    subgraph: subgraphName,
                                    value: example,
                                },
                            });
                        }
                        inputTypeComposer.setDirectives(directives);
                    }
                    inputTypeComposer.setExtensions({
                        default: subSchemaAndTypeComposers.default,
                    });
                }
                else {
                    subSchemaAndTypeComposers.input = schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON);
                }
            }
            switch (subSchemaAndTypeComposers.type) {
                case 'object':
                    if (subSchemaAndTypeComposers.properties) {
                        for (const propertyName in subSchemaAndTypeComposers.properties) {
                            // TODO: needs to be fixed
                            if (propertyName === 'additionalProperties') {
                                continue;
                            }
                            const fieldName = (0, utils_1.sanitizeNameForGraphQL)(propertyName);
                            const fieldDirectives = [];
                            if (propertyName !== fieldName) {
                                schemaComposer.addDirective(directives_js_1.ResolveRootFieldDirective);
                                fieldDirectives.push({
                                    name: 'resolveRootField',
                                    args: {
                                        subgraph: subgraphName,
                                        field: propertyName,
                                    },
                                });
                            }
                            if (subSchemaAndTypeComposers.properties[propertyName].flatten) {
                                schemaComposer.addDirective(directives_js_1.FlattenDirective);
                                fieldDirectives.push({
                                    name: 'flatten',
                                    args: {
                                        subgraph: subgraphName,
                                    },
                                });
                            }
                            fieldMap[fieldName] = {
                                type: () => {
                                    const typeComposers = subSchemaAndTypeComposers.properties[propertyName];
                                    let nullable = true;
                                    if (subSchemaAndTypeComposers.required?.includes(propertyName)) {
                                        nullable = false;
                                    }
                                    // Nullable has more priority
                                    if (typeComposers.nullable === false) {
                                        nullable = false;
                                    }
                                    if (typeComposers.nullable === true) {
                                        nullable = true;
                                    }
                                    if (subSchemaAndTypeComposers.properties[propertyName].writeOnly) {
                                        nullable = true;
                                    }
                                    return !nullable ? typeComposers.output.getTypeNonNull() : typeComposers.output;
                                },
                                // Make sure you get the right property
                                directives: fieldDirectives,
                                extensions: {
                                    nullable: subSchemaAndTypeComposers.properties?.[propertyName]?.nullable,
                                },
                                description: subSchemaAndTypeComposers.properties[propertyName].description ||
                                    subSchemaAndTypeComposers.properties[propertyName].output?.description,
                                deprecationReason: subSchemaAndTypeComposers.properties[propertyName].deprecated
                                    ? 'deprecated'
                                    : undefined,
                            };
                            const directives = [];
                            if (fieldName !== propertyName) {
                                schemaComposer.addDirective(directives_js_1.ResolveRootFieldDirective);
                                directives.push({
                                    name: 'resolveRootField',
                                    args: {
                                        subgraph: subgraphName,
                                        field: propertyName,
                                    },
                                });
                            }
                            inputFieldMap[fieldName] = {
                                type: () => {
                                    const typeComposers = subSchemaAndTypeComposers.properties[propertyName];
                                    let nullable = true;
                                    if (subSchemaAndTypeComposers.required?.includes(propertyName)) {
                                        nullable = false;
                                    }
                                    // Nullable has more priority
                                    if (typeComposers.nullable === false) {
                                        nullable = false;
                                    }
                                    if (typeComposers.nullable === true) {
                                        nullable = true;
                                    }
                                    if (subSchemaAndTypeComposers.properties[propertyName].readOnly) {
                                        nullable = true;
                                    }
                                    return !nullable ? typeComposers.input?.getTypeNonNull() : typeComposers.input;
                                },
                                extensions: {
                                    nullable: subSchemaAndTypeComposers.properties?.[propertyName]?.nullable,
                                },
                                directives,
                                description: subSchemaAndTypeComposers.properties[propertyName].description ||
                                    subSchemaAndTypeComposers.properties[propertyName].input?.description,
                                defaultValue: subSchemaAndTypeComposers.properties[propertyName]?.default ||
                                    subSchemaAndTypeComposers.properties[propertyName]?.extensions?.default ||
                                    subSchemaAndTypeComposers.properties[propertyName]?.input?.default,
                                // deprecationReason: subSchemaAndTypeComposers.properties[propertyName].deprecated ? 'deprecated' : undefined,
                            };
                        }
                    }
                    if (subSchemaAndTypeComposers.additionalProperties) {
                        // Take a look later
                        if (typeof subSchemaAndTypeComposers.additionalProperties === 'object' &&
                            subSchemaAndTypeComposers.additionalProperties.output instanceof graphql_compose_1.ObjectTypeComposer) {
                            const containerOutputTC = schemaComposer.createObjectTC({
                                name: `${subSchemaAndTypeComposers.additionalProperties.output.getTypeName()}_entry`,
                                fields: {
                                    key: {
                                        type: 'ID!',
                                    },
                                    value: {
                                        type: subSchemaAndTypeComposers.additionalProperties.output,
                                    },
                                },
                            });
                            schemaComposer.addDirective(directives_js_1.DictionaryDirective);
                            fieldMap.additionalProperties = {
                                type: containerOutputTC.List,
                                directives: [
                                    {
                                        name: 'dictionary',
                                        args: {
                                            subgraph: subgraphName,
                                        },
                                    },
                                ],
                            };
                        }
                        else if (Object.keys(fieldMap).length > 0) {
                            schemaComposer.addDirective(directives_js_1.ResolveRootDirective);
                            fieldMap.additionalProperties = {
                                type: graphql_scalars_1.GraphQLJSON,
                                directives: [
                                    {
                                        name: 'resolveRoot',
                                        args: {
                                            subgraph: subgraphName,
                                        },
                                    },
                                ],
                            };
                        }
                        else {
                            const typeComposer = schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON);
                            schemaComposer.delete(subSchemaAndTypeComposers.input?.getTypeName?.());
                            schemaComposer.delete(subSchemaAndTypeComposers.output?.getTypeName?.());
                            return {
                                input: typeComposer,
                                output: typeComposer,
                                description: subSchemaAndTypeComposers.description,
                                nullable: subSchemaAndTypeComposers.nullable,
                                default: subSchemaAndTypeComposers.default,
                                readOnly: subSchemaAndTypeComposers.readOnly,
                                writeOnly: subSchemaAndTypeComposers.writeOnly,
                                deprecated: subSchemaAndTypeComposers.deprecated,
                            };
                        }
                    }
                    if (subSchemaAndTypeComposers.title in rootInputTypeNameComposerMap) {
                        const typeComposer = rootInputTypeNameComposerMap[subSchemaAndTypeComposers.title]();
                        for (const fieldName in inputFieldMap) {
                            let inputTC = inputFieldMap[fieldName].type();
                            if ('ofType' in inputTC) {
                                inputTC = inputTC.ofType;
                            }
                            typeComposer.addFieldArgs(fieldName, inputTC.getFields());
                        }
                        return {
                            output: typeComposer,
                        };
                    }
                    let output = subSchemaAndTypeComposers.output;
                    if (Object.keys(fieldMap).length === 0) {
                        output = schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON);
                    }
                    else if ('addFields' in output) {
                        if (subSchemaOnly.discriminatorMapping) {
                            for (const discriminatorValue in subSchemaOnly.discriminatorMapping) {
                                const discType = subSchemaOnly.discriminatorMapping[discriminatorValue];
                                discType.output.addFields(fieldMap);
                            }
                        }
                        output.addFields(fieldMap);
                        // TODO: Improve this later
                        for (const requiredFieldName of subSchemaAndTypeComposers.required || []) {
                            const sanitizedFieldName = (0, utils_1.sanitizeNameForGraphQL)(requiredFieldName);
                            const fieldObj = output.getField(sanitizedFieldName);
                            if (!(0, graphql_1.isNonNullType)(fieldObj.type.getType()) && !fieldObj.extensions?.nullable) {
                                output.makeFieldNonNull(requiredFieldName);
                            }
                        }
                    }
                    let input = subSchemaAndTypeComposers.input;
                    if (Object.keys(inputFieldMap).length === 0) {
                        input = schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON);
                    }
                    else if (input != null && 'addFields' in input) {
                        input.addFields(inputFieldMap);
                        // TODO: Improve this later
                        for (const requiredFieldName of subSchemaAndTypeComposers.required || []) {
                            const sanitizedFieldName = (0, utils_1.sanitizeNameForGraphQL)(requiredFieldName);
                            const field = input.getField(sanitizedFieldName);
                            if (!(0, graphql_1.isNonNullType)(field.type.getType()) && !field.extensions?.nullable) {
                                input.makeFieldNonNull(requiredFieldName);
                            }
                        }
                    }
                    if (isList) {
                        input = input.List;
                        output = output.List;
                    }
                    return {
                        input,
                        output,
                        nullable: subSchemaAndTypeComposers.nullable,
                        default: subSchemaAndTypeComposers.default,
                        readOnly: subSchemaAndTypeComposers.readOnly,
                        writeOnly: subSchemaAndTypeComposers.writeOnly,
                        deprecated: subSchemaAndTypeComposers.deprecated,
                    };
            }
            if (subSchemaAndTypeComposers.input || subSchemaAndTypeComposers.output) {
                return {
                    input: subSchemaAndTypeComposers.input,
                    output: subSchemaAndTypeComposers.output,
                    description: subSchemaAndTypeComposers.description,
                    nullable: subSchemaAndTypeComposers.nullable,
                    default: subSchemaAndTypeComposers.default,
                    readOnly: subSchemaAndTypeComposers.readOnly,
                    writeOnly: subSchemaAndTypeComposers.writeOnly,
                    deprecated: subSchemaAndTypeComposers.deprecated,
                };
            }
            else {
                logger.debug(`GraphQL Type cannot be created for this JSON Schema definition;`, {
                    subSchema: subSchemaOnly,
                    path,
                });
                const typeComposer = schemaComposer.getAnyTC(graphql_scalars_1.GraphQLJSON);
                return {
                    input: typeComposer,
                    output: typeComposer,
                    description: subSchemaAndTypeComposers.description,
                    nullable: subSchemaAndTypeComposers.nullable,
                    readOnly: subSchemaAndTypeComposers.readOnly,
                    writeOnly: subSchemaAndTypeComposers.writeOnly,
                    default: subSchemaAndTypeComposers.default,
                    deprecated: subSchemaAndTypeComposers.deprecated,
                };
            }
        },
    });
}
const specifiedTypeNames = ['String', 'Int', 'Float', 'Boolean', 'ID', 'JSON', 'Void'];
function isUnspecificType(typeComposer) {
    const tc = typeComposer.getTypeName();
    return !specifiedTypeNames.includes(tc);
}
